# frozen_string_literal: true

# Copyright from 2019 Paulo Arruda (parrudaj at gmail dot com)
#
# Licensed under the Commons Clause + Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

module Gotta
  module Serve
    class App < Sinatra::Application
      set :show_exceptions, SHOW_EXCEPTIONS
      set :logging, LOGGING
      set :root, SERVER_ROOT
      set :public_folder, PUBLIC_DIR
      set :static, SERVE_STATIC
      set :static_cache_control, [:must_revalidate, :proxy_revalidate, max_age: 0]

      register Sinatra::MultiRoute

      before do
        cache_control :must_revalidate, :proxy_revalidate, max_age: 0
      end

      route :head, :get, :post, :put, :patch, :delete, '/*' do
        splat = params['splat'][0]
        component_name = resolve_component_name(splat)
        request_headers, original_request_id = set_request_headers(env)

        body = parse_body(request,
                          request_headers['Content-Type'],
                          request.request_method)

        query_params = parse_query(request.query_string)

        secrets = Gotta::Project.config.secrets[ENVIRONMENT]
        context = { environment: ENVIRONMENT }

        event = Gotta::Event.new(body: body,
                                 query_params: query_params,
                                 headers: request_headers,
                                 context: context,
                                 secrets: secrets)

        log_request_message(component_name,
                            request,
                            request_uuid,
                            query_params,
                            body,
                            request_headers)

        payload = {
          load_path: component_name,
          response_queue: SecureRandom.uuid
          event: event
        }
        queue = Gotta::Serve.input_queue.push(payload.to_msgpack)
        response_queue = Gotta::Serve.queue_driver.new(payload[:response_queue])

        time, response = parse_response(response_queue.pop)

        status response.status
        headers set_response_headers(response,
                                     request_uuid,
                                     original_request_id,
                                     time)

        response_body, print_body = parse_response(response)
        log_response_message(component_name,
                             time,
                             request_uuid,
                             response,
                             print_body)

        body response_body
      end

      def set_request_headers(env)
        request_headers = parse_headers(env)
        request_headers['X-Request-Id'] = SecureRandom.uuid
        request_headers['Request-Method'] = request.request_method
        [request_headers, request_headers['X-Original-Request-Id']]
      end

      def log_request_message(component_name,
                              request,
                              request_uuid,
                              query_params,
                              body,
                              request_headers)

        puts "[#{component_name}] <- [REQUEST: #{request.request_method} \"#{request.fullpath}\"] request_id=\"#{request_uuid}\" body=#{body.inspect} query_params=#{query_params} headers=#{request_headers}"
      end

      def log_response_message(component_name,
                               time,
                               request_uuid,
                               response,
                               print_body)

        puts "[#{component_name}] -> [RESPONSE: #{time}ms] request_id=\"#{request_uuid}\" status=#{response.status} body=#{print_body.inspect} headers=#{response.headers}"
      end

      def set_response_headers(response,
                               request_uuid,
                               original_request_id = nil,
                               time)

        response.headers['X-Request-Id'] = request_uuid
        response.headers['X-Original-Request-Id'] = original_request_id if original_request_id
        response.headers['X-Execution-time'] = "#{time}ms"
        response.headers
      end

      def parse_response(response)
        return [Base64.urlsafe_decode64(response.body), "Base64(#{response.body})"] if response.binary?

        [response.body, response.body]
      end

      def resolve_component_name(splat)
        return Gotta::Project.config.root_to if splat == ''
        return Gotta::Project.config.catch_all unless File.directory?(name)

        # TODO: 'catch_all' should render when the handler file
        # cannot be found.
        # Maybe this could happen when the handler file can't
        # be found (trigger by gotta-mod)

        # if !is_a_component?(splat)
        #   return Gotta::Project.config.catch_all
        # end
        splat
      end

      def parse_body(request, content_type, method)
        return nil if method == 'GET'
        return form_content(request.env) if content_type == 'application/x-www-form-urlencoded'

        body_content(request.body)
      end

      def form_content(request_env)
        return {} if (content = request_env['rack.request.form_hash']).nil?

        content.to_json
      end

      def body_content(request_body)
        return {} if (content = request_body.read).nil?

        content
      end

      def parse_query(query_string)
        hash = {}
        query_string.split('&').each do |param|
          key, value = param.split('=')
          hash[key] = value
        end
        hash
      end

      def parse_headers(env)
        result = {}
        env.select { |e| e.match(/^HTTP_/) }.each do |k, v|
          newkey = k.sub(/^HTTP_/, '').split('_').map(&:capitalize).join('-')
          result[newkey] = v
        end
        result['Content-Type'] = env['CONTENT_TYPE']
        result['Request-Method'] = env['REQUEST_METHOD']
        result['Content-Length'] = env['CONTENT_LENGTH']
        result['Remote-Addr'] = env['REMOTE_ADDR']
        result
      end
    end
  end
end
